--- old/usr/src/uts/common/fs/zfs/dmu.c	2013-05-01 11:13:43.600935029 +0200
+++ new/usr/src/uts/common/fs/zfs/dmu.c	2013-05-01 11:13:43.111401819 +0200
@@ -21,6 +21,7 @@
 /*
  * Copyright (c) 2005, 2010, Oracle and/or its affiliates. All rights reserved.
  * Copyright (c) 2013 by Delphix. All rights reserved.
+ * Copyright (c) 2013 Martin Matuska. All rights reserved.
  */
 
 #include <sys/dmu.h>
@@ -42,6 +43,7 @@
 #include <sys/zio_checksum.h>
 #include <sys/zio_compress.h>
 #include <sys/sa.h>
+#include <sys/zfeature.h>
 #ifdef _KERNEL
 #include <sys/vmsystm.h>
 #include <sys/zfs_znode.h>
@@ -1417,7 +1419,7 @@
 
 	DB_DNODE_ENTER(db);
 	dn = DB_DNODE(db);
-	dmu_write_policy(os, dn, db->db_level, WP_DMU_SYNC, &zp);
+	dmu_write_policy(os, dn, db->db_level, WP_DMU_SYNC, &zp, txg);
 	DB_DNODE_EXIT(db);
 
 	/*
@@ -1555,9 +1557,11 @@
 }
 
 int zfs_mdcomp_disable = 0;
+int zfs_mdcomp_lz4 = 0;
 
 void
-dmu_write_policy(objset_t *os, dnode_t *dn, int level, int wp, zio_prop_t *zp)
+dmu_write_policy(objset_t *os, dnode_t *dn, int level, int wp, zio_prop_t *zp,
+    uint64_t txg)
 {
 	dmu_object_type_t type = dn ? dn->dn_type : DMU_OT_OBJSET;
 	boolean_t ismd = (level > 0 || DMU_OT_IS_METADATA(type) ||
@@ -1582,8 +1586,26 @@
 		 * XXX -- we should design a compression algorithm
 		 * that specializes in arrays of bps.
 		 */
-		compress = zfs_mdcomp_disable ? ZIO_COMPRESS_EMPTY :
-		    ZIO_COMPRESS_LZJB;
+		if (zfs_mdcomp_disable)
+			compress = ZIO_COMPRESS_EMPTY;
+		else if (zfs_mdcomp_lz4 && os->os_spa != NULL) {
+			zfeature_info_t *feat = &spa_feature_table
+			    [SPA_FEATURE_LZ4_COMPRESS];
+
+			if (spa_feature_is_active(os->os_spa, feat))
+				compress = ZIO_COMPRESS_LZ4;
+			else if (spa_feature_is_enabled(os->os_spa, feat)) {
+				dmu_tx_t *tx;
+
+				tx = dmu_tx_create_assigned(
+				    spa_get_dsl(os->os_spa), txg);
+				spa_feature_incr(os->os_spa, feat, tx);
+				dmu_tx_commit(tx);
+				compress = ZIO_COMPRESS_LZ4;
+			} else
+				compress = ZIO_COMPRESS_LZJB;
+		} else
+			compress = ZIO_COMPRESS_LZJB;
 
 		/*
 		 * Metadata always gets checksummed.  If the data
